//###########################################################
// File: compact.c
//
// Read-/Writeroutines for CompactFlash in LBA mode.
// CHS mode not supported.
//
//#########################################################################
// Last change: 06.03.2004
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################
// Compiler: AVR-GCC 3.2
//#########################################################################

#include <io.h>

#include "mydefs.h"
#include "protos.h"
#include "dos.h"
#include "serial.h"

unsigned long maxsect;

unsigned char inbuff[BYTE_PER_SEC];  //buffer for FAT- and dirsectors
unsigned char iob[BYTE_PER_SEC];     //file i/o buffer

//######################################################
unsigned char CFReadSector(unsigned long lba, unsigned char *buf)
//######################################################
{
 unsigned int i;
 unsigned char by;
 unsigned char *p;

 if(lba>=maxsect) return 1; //sectornumber too big

 if(CFWaitReady()) return 1;
 
// CFWriteAdr(CF_FEATURES,0); //Brauch ich nicht
 CFWriteAdr(CF_SECCOUNT,1);   //einen Sektor lesen

 by=(unsigned char)lba;
 CFWriteAdr(CF_LBA0,by);            //D7..0
 by=(unsigned char)(lba>>8);
 CFWriteAdr(CF_LBA1,by);            //D15..8
 by=(unsigned char)(lba>>16);
 CFWriteAdr(CF_LBA2,by);            //D23..16
 by=(unsigned char)(lba>>24);
 by&=0x0F;              //Nur die vier unteren Bits gltig
 by|=0xE0;              //LBA-Mode Drive0
 CFWriteAdr(CF_LBA3,by);          //D27..24
 CFWriteAdr(CF_STACOM,CF_READ_SEC);

 if(CFWaitDrq()) return 2;

 p=buf;
 
 CF_DATA_DIR_IN();
 CFSetAdr(CF_IO);
 CF_CS_OFF();
 for(i=0; i<BYTE_PER_SEC; i++)
  {
   CF_RD_OFF();
   NOP(); //Schon bei 8MHz mu man hier etwas warten
   NOP(); //Es knnte evtl. ntig sein hier noch mehr
        //nop einzufgen, um mglichst viele CF zu lesen
        //Bei mir reichten bisher 2 nop
        //Sicherheitshalber noch einen mehr
   NOP();

#ifdef F_16 //16MHz processor clock
//   NOP();
//   NOP();
   NOP();
#endif
   
   *p++=CF_READ_DATA();
   CF_RD_ON();
  } 

 CF_CS_ON();

 return 0;
}

//######################################################
unsigned char CFWriteSector(unsigned long lba, unsigned char *buf)
//######################################################
{
 unsigned int i;
 unsigned char by;
 unsigned char *p;

 if(lba>=maxsect) return 1; //sectornumber too big
 if(CFWaitReady()) return 1;
 
// CFWriteAdr(CF_FEATURES,0); //Brauch ich nicht
 CFWriteAdr(CF_SECCOUNT,1);   //einen Sektor schreiben

 by=(unsigned char)lba;
 CFWriteAdr(CF_LBA0,by);            //D7..0
 by=(unsigned char)(lba>>8);
 CFWriteAdr(CF_LBA1,by);            //D15..8
 by=(unsigned char)(lba>>16);
 CFWriteAdr(CF_LBA2,by);            //D23..16
 by=(unsigned char)(lba>>24);
 by&=0x0F;              //Nur die vier unteren Bits gltig
 by|=0xE0;              //LBA-Mode Drive0
 CFWriteAdr(CF_LBA3,by);          //D27..24
 CFWriteAdr(CF_STACOM,CF_WRITE_SEC);

 if(CFWaitDrq()) return 2;

 p=buf; //using a pointer is much faster than indexing inbuff[i]

 CFSetAdr(CF_IO);
 CF_DATA_DIR_OUT();
 CF_CS_OFF();

 for(i=0; i<BYTE_PER_SEC; i++)
  {
//   CF_WRITE_DATA(inbuff[i]);
   CF_WRITE_DATA(*p++);
#ifdef F_16 //16MHz processor clock
   NOP();
#endif
   CF_WR_OFF();
   NOP();
#ifdef F_16 //16MHz processor clock
   NOP();
#endif
   
   CF_WR_ON();
   NOP();
#ifdef F_16 //16MHz processor clock
   NOP();
#endif
  }

 CF_CS_ON();

 CF_DATA_DIR_IN();
 return 0;
}

//######################################################
unsigned char CFIdentify(void)
//######################################################
{
 unsigned long maxtracks;
 unsigned int heads,sectors_per_track;
 unsigned int i;
 union Convert *cv;

 if(CFWaitReady()) return 1;
 
// cf_features=0; //Brauch ich nicht
// cf_seccount=1;   //einen Sektor lesen
// cf_lba0=1;       //LBA ist hier eigentlich egal, also 1
// cf_lba1=0;
// cf_lba2=0;
// cf_lba3=0xE0;
 CFWriteAdr(CF_STACOM,CF_IDENTIFY);

 if(CFWaitDrq()) return 2;
  
 for(i=0; i<BYTE_PER_SEC; i++) inbuff[i]=CFReadAdr(CF_IO);
 
 cv=(union Convert *)&inbuff[2];
// maxtracks=MakeULONG(&inbuff[2]);          //Anzahl Tracks
 maxtracks=cv->ul;          //Anzahl Tracks

 cv=(union Convert *)&inbuff[6];
// heads=MakeUINT(&inbuff[6]);              //Anzahl Heads
 heads=cv->ui;              //Anzahl Heads

 cv=(union Convert *)&inbuff[12];
 sectors_per_track=cv->ui; //Sectors per Track

 cv=(union Convert *)&inbuff[14];
 maxsect=(unsigned long)cv->ui << 16;         //Anzahl Sektoren
 cv=(union Convert *)&inbuff[16];
 maxsect+=cv->ui;

//Info's zum Laufwerk anzeigen
 
#ifdef F_DEBUG
 printf("Drive Name: ");
 for(i=54; i<94; i+=2) //Hersteller String 
  {
   putchar(inbuff[i+1]);
   putchar(inbuff[i]);
  }
 printf("\n");

 printf("Revision: ");
 for(i=46; i<54; i+=2) //Revisions String
  {
   putchar(inbuff[i+1]);
   putchar(inbuff[i]);
  }

 printf("\nSerialNumber: ");
 for(i=20; i<20+20; i++) putchar(inbuff[i]); //SerialNumber String

 printf("\nC/H/S %lu/%u/%lu\n",maxtracks,heads,maxsect);
 printf("Capacity: %lukB\n",maxsect/2);
#endif

 return 0;
}

//######################################################
unsigned char CFWaitReady(void)
//######################################################
{
 unsigned char by;

 by=CFReadAdr(CF_STACOM);
 if(by & 0x01) return 1; //Fehler !

 do
  {
   by=CFReadAdr(CF_STACOM);
   by&=0xF0;
   
  }while(by!=0x50); //Endlosschleife mglich !
  
 return 0; 
}

//######################################################
unsigned char CFWaitDrq(void)
//######################################################
{
 unsigned char by;

 by=CFReadAdr(CF_STACOM);
 if(by & 0x01) return 1; //Fehler !

 do
  {
   by=CFReadAdr(CF_STACOM);
   by&=0xF8;
   
  }while(by!=0x58); //Endlosschleife mglich !
  
 return 0; 
}

//######################################################
unsigned char CFReadAdr(unsigned char adr)
//######################################################
{
 unsigned char by;
 
 CF_DATA_DIR_IN();
 CFSetAdr(adr);
 CF_CS_OFF();
 CF_RD_OFF();
 NOP(); //Schon bei 8MHz mu man hier etwas warten
 NOP(); //Es knnte evtl. ntig sein hier noch mehr
        //nop einzufgen, um mglichst viele CF zu lesen
        //Bei mir reichten bisher 2 nop
        //Sicherheitshalber noch einen mehr
 NOP();
        //Sonst meldet CFWaitReady() immer Fehler

#ifdef F_16 //16MHz processor clock
//   NOP();
//   NOP();
   NOP();
#endif
   
 by=CF_READ_DATA();
 CF_RD_ON();
 CF_CS_ON();
 return by;
}

//######################################################
unsigned char CFRead(void)
//######################################################
{
 unsigned char by;
 
 CF_RD_OFF();
 NOP(); //Schon bei 8MHz mu man hier etwas warten
 NOP(); //Es knnte evtl. ntig sein hier noch mehr
        //nop einzufgen, um mglichst viele CF zu lesen
        //Bei mir reichten bisher 2 nop
        //Sicherheitshalber noch einen mehr
 NOP();

#ifdef F_16 //16MHz processor clock
//   NOP();
//   NOP();
   NOP();
#endif

 by=CF_READ_DATA();
 CF_RD_ON();
 return by;
}

//######################################################
void CFWriteAdr(unsigned char adr, unsigned char dat)
//######################################################
{
 CF_DATA_DIR_OUT();
 CFSetAdr(adr);
 CF_WRITE_DATA(dat);
 NOP();
#ifdef F_16 //16MHz processor clock
   NOP();
#endif

 CF_CS_OFF();
 NOP();
#ifdef F_16 //16MHz processor clock
   NOP();
#endif
 CF_WR_OFF();
 NOP();
#ifdef F_16 //16MHz processor clock
   NOP();
#endif
 CF_WR_ON();
 CF_CS_ON();
}

//######################################################
//Ist so kurz das man statt der Routine auch direkt oben
//per Macro einfgen knnte. Wird dann etwas schneller.
void CFSetAdr(unsigned char adr)
//######################################################
{
 unsigned char by;

 by=CF_ADR_PIN;  //Portpins lesen
 by&=0xF0;       //D0..D3 lschen
 by|=adr;        //adr reinodern
 CF_ADR_PORT=by; //und ausgeben
}

/*
//###########################################################
//Macht aus 4 Bytes vom InputBuffer einen ULONG Wert
//Achtung: Nur fr CFIdentify() ! Reihenfolge LOW-HIGH !
unsigned long MakeULONG1(unsigned char *dat)
//###########################################################
{
 unsigned long tmp;

 tmp=(unsigned long)(dat[0])<<16;
 tmp+=(unsigned long)(dat[1])<<24;
 tmp+=dat[2];
 tmp+=(unsigned long)(dat[3])<<8;

 return tmp;
}
*/
/*
//###########################################################
//Macht aus 4 Bytes vom InputBuffer einen ULONG Wert
unsigned long MakeULONG(unsigned char *dat)
//###########################################################
{
 unsigned long tmp;

 tmp=dat[0];
 tmp+=(unsigned long)(dat[1])<<8;
 tmp+=(unsigned long)(dat[2])<<16;
 tmp+=(unsigned long)(dat[3])<<24;

 return tmp;
}
*/
